
/*

File: noode.js

Based on John Resig Simple JavaScript Inheritance: http://ejohn.org/ with some nice additions

Check out http://ejohn.org/blog/simple-javascript-inheritance/ 
for discussion about the first implementation

Both John Resig work and Noode.js are MIT licensed.


	
*/


var initializing = false, 
	fnTest = /xyz/.test(function(){xyz;}) ? /\b_super\b/ : /.*/,
/*
  	Class: Class
  	
  	The base class
*/
	Class = function () {};

/*
  	Function: Class.extend
  	
  	Create a new class that inherits from the base Class object
  	
  	Parameters:
  		prop - [object] Methods to add to the new class
  		
  	Returns:
  	The subclass
*/
Class.extend = function(prop)
{
	var _super = this.prototype,
		name ;
	
	// Instantiate a base class (but only create the instance,
	// don't run the init constructor)
	initializing = true;
	
	prototype = new this();
	
	initializing = false;
	
	// Copy the properties over onto the new prototype
	for (name in prop)
	{
		
		// Check if we're overwriting an existing function
		prototype[name] = typeof prop[name] == "function" && 
		typeof _super[name] == "function" && fnTest.test(prop[name]) ?
				
			(function(name, fn){
				return function() {
					var tmp = this._super,
						ret ;
					
					// Add a new ._super() method that is the same method
					// but on the super-class
					this._super = _super[name];
					
					// The method only need to be bound temporarily, so we
					// remove it when we're done executing
					ret = fn.apply(this, arguments);    
					
					this._super = tmp;
					
					return ret;
				};
			})(name, prop[name]) :
				
			prop[name];
	}
	
	// The dummy class constructor
	function Class() {
		// All construction is actually done in the init method
		if ( !initializing )
		{
			if ( typeof this.construct == 'function' ) {
				this.construct.apply(this, arguments);
			}
		} 
	}
	
	// Populate our constructed prototype object
	Class.prototype = prototype;

	/*
		Function: delegate
		
		Method for easy delegation to the current instance
		
		Parameters:
			fn - [string] The name of the function 
		
	 */
	Class.prototype.delegate = function ( fn )
	{
		var a = this[fn],
			b = this ;
		
		return function () {
			context.__caller = this ;
			return a.apply ( b , arguments );
		};
	};
	
	// Enforce the constructor to be what we expect
	Class.constructor = Class;
	
	// And make this class extendable
	Class.extend = arguments.callee;
	
	return Class;
};


/*
	Class: Event
	
*/
var Event = Class.extend({
	/*
		Constructor: construct
		
		Parameters:
			name - [string] Name of the event to dispatch
			parameters - [object] Parameters of the new event
			
	*/
	construct: function ( name , parameters )
	{
		if ( typeof(name) != 'string' )
		{
			throw 'Noode - Event - name param must be a string' ;
		}
		if ( parameters && typeof(parameters) != 'object' )
		{
			throw 'Noode - Event - parameters param must be an object' ;
		}

		var p = parameters || {} ;
		
		for ( k in p )
		{
			this[k] = p[k] ;
		}

		this.prevented_ = false ;
		
		this.name_ = name ;
		
	},

	/*
		Function: getName
		
		Get the name of the event
		
		Returns:
		The name of the event
	*/
	getName: function ()
	{
		return this.name_ ;
	},

	/*
		Function: prevent
		
		Prevent the default action of the event
	*/
	prevent: function ()
	{
		this.prevented_ = true ;
	},

	/*
		Function: isPrevented
		
		Get current prevented state of the event
		
		Returns:
		True is the event has been prevented by a listener, false otherwise
	*/
	isPrevented: function ()
	{
		return this.prevented_ ;
	}
});

/*
	Class: AbstractEventDispatcher
	
*/
var AbstractEventDispatcher = Class.extend({
	
	
	/*
		Function: dispatch
		
		Parameters:
			event - [Event|string] Name of the event to dispatch
			parameters - [object] Event parameters object to dispatch to listeners, optional, only if a string given as first parameters
			
		Returns:
		False if a listener has returned false (event has been prevented), true otherwise 
	*/
	dispatch: function ( event , parameters )
	{

		// There is no listener, so no need to send any event
		if (!this.listeners__)
		{
			return true;
		}
		
		var evt ;
		
		if ( typeof(event) == 'string' )
		{
			evt = new Event( event , parameters ) ;
		} else if ( !(event instanceof Event) )
		{
			throw 'Noode - AbstractEventDispatcher - event param must be a string or an Event object' ;
		} else {
			evt = event ;
		}


		var ret = true,
			k,
			name = evt.getName() ;
	
		if ( typeof(this.listeners__[name])=='object' )
		{
			for ( k in this.listeners__[name] )
			{
				this.listeners__[name][k] ( evt ) ;
			}
		}
	
		return evt.isPrevented () === false ;
	},
	/*
		Function: on
		
		Adds an event listener
		
		Parameters:
			name - [string] Name of the event to listen
			callback - [function] Function to call when event is dispatched
			
		Returns:
		Current instance for chained commands on this element
	*/
	on: function ( name , callback )
	{
		if (!this.listeners__)
		{
			this.listeners__ = {} ;
		}
		
		if ( typeof(name) != 'string' )
		{
			throw 'Noode - AbstractEventDispatcher - name param must be a string' ;
		}
		if ( typeof(callback) != 'function' )
		{
			throw 'Noode - AbstractEventDispatcher - callback param must be a function' ;
		}
		
		if ( typeof(this.listeners__[name]) != 'object' )
		{
			this.listeners__[name] = [] ;
		}
		
		this.listeners__[name].push(callback);
		
		return this;
	},
	/*
		Function: addListener
		
		Alias of AbstractEventDispatcher.on
	*/
	addListener: function ( name, callback ) 
	{
		return this.on(name, callback);
	},
	/*
		Function: off
		
		Removes an event listener
		
		Parameters:
			name - [string] Name of the event to listen
			callback - [function] Function to call when event is dispatched
			
		Returns:
		Current instance for chained commands on this element
	*/
	off: function ( name , callback )
	{
		if (!this.listeners__)
		{
			return; ;
		}
		
		if ( typeof(name) != 'string' )
		{
			throw 'Noode - AbstractEventDispatcher - event param must be a string' ;
		}
		if ( typeof(callback) != 'function' )
		{
			throw 'Noode - AbstractEventDispatcher - callback param must be a function' ;
		}
		
		if ( typeof(this.listeners__[name]) == 'object' )
		{
			for ( var k in this.listeners__[name] )
			{
				if ( this.listeners__[name][k] == callback )
				{
					delete(this.listeners__[name][k]);
					break;
				}
			}
		}
		
		return this;
	},
	/*
		Function: remListener
		
		Alias of AbstractEventDispatcher.off
	*/
	remListener: function ( name, callback ) 
	{
		return this.off(name, callback);
	}

});



exports.Event = Event ;
exports.Class = Class ;
exports.AbstractEventDispatcher = AbstractEventDispatcher ;
